package net.black3r.bakalarka.bakalarka;

import android.util.Log;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;
import java.io.UnsupportedEncodingException;
import java.security.*;

public class AESCrypter implements Crypter {
    private byte[] key_bytes;
    private SecretKeySpec key;
    private SecureRandom rng;

    @Override
    public IvParameterSpec generateIv() {
        byte[] iv_bytes = new byte[16];
        rng.nextBytes(iv_bytes);
        return new IvParameterSpec(iv_bytes);
    }

    public static AESCrypter fromBase64(String key) {
        return new AESCrypter(Base64Wrap.decode(key));
    }

    public AESCrypter(String key) {
        rng = new SecureRandom();
        try {
            this.key_bytes = key.getBytes("UTF-8");
        } catch (UnsupportedEncodingException e) {
            e.printStackTrace();
        }
        this.make_key();
    }

    public AESCrypter(byte[] key) {
        rng = new SecureRandom();
        this.key_bytes = key;
        this.make_key();
    }

    private void make_key() {
        this.key = new SecretKeySpec(this.key_bytes, "AES");
    }

    @Override
    public byte[] encryptToByteArray(byte[] input) {
        IvParameterSpec iv = generateIv();
        return encryptToByteArray(input, iv);
    }

    @Override
    public byte[] encryptToByteArray(String input) {
        IvParameterSpec iv = generateIv();
        return encryptToByteArray(input, iv);
    }

    @Override
    public String encryptToBase64(byte[] input) {
        IvParameterSpec iv = generateIv();
        return encryptToBase64(input, iv);
    }

    @Override
    public String encryptToBase64(String input) {
        IvParameterSpec iv = generateIv();
        return encryptToBase64(input, iv);
    }

    @Override
    public byte[] encryptToByteArray(byte[] input, IvParameterSpec iv) {
        Cipher cipher = null;
        try {
            cipher = Cipher.getInstance("AES/CBC/PKCS5Padding");
        } catch (NoSuchAlgorithmException e) {
            e.printStackTrace();
        } catch (NoSuchPaddingException e) {
            e.printStackTrace();
        }
        try {
            cipher.init(Cipher.ENCRYPT_MODE, key, iv);
        } catch (InvalidKeyException e) {
            e.printStackTrace();
        } catch (InvalidAlgorithmParameterException e) {
            e.printStackTrace();
        }
        try {
            byte[] ciphertext = cipher.doFinal(input);
            byte[] iv_bytes = iv.getIV();
            byte[] result = new byte[iv_bytes.length + ciphertext.length];
            for (int i = 0; i < 16; i++) {
                result[i] = iv_bytes[i];
            }
            for (int i = 16; i < result.length; i++) {
                result[i] = ciphertext[i-16];
            }
            return result;
        } catch (IllegalBlockSizeException e) {
            e.printStackTrace();
        } catch (BadPaddingException e) {
            e.printStackTrace();
        }
        return null;
    }

    @Override
    public byte[] encryptToByteArray(String input, IvParameterSpec iv) {
        try {
            return encryptToByteArray(input.getBytes("UTF-8"), iv);
        } catch (UnsupportedEncodingException e) {
            e.printStackTrace();
        }
        return null;
    }

    @Override
    public byte[] decryptFromByteArray(byte[] input, IvParameterSpec iv) {
        Cipher cipher = null;
        try {
            cipher = Cipher.getInstance("AES/CBC/PKCS5Padding");
        } catch (NoSuchAlgorithmException e) {
            e.printStackTrace();
        } catch (NoSuchPaddingException e) {
            e.printStackTrace();
        }
        try {
            cipher.init(Cipher.DECRYPT_MODE, key, iv);
        } catch (InvalidKeyException e) {
            e.printStackTrace();
        } catch (InvalidAlgorithmParameterException e) {
            e.printStackTrace();
        }
        try {
            return cipher.doFinal(input);
        } catch (IllegalBlockSizeException e) {
            e.printStackTrace();
        } catch (BadPaddingException e) {
            // TODO this is caught when invalid key is used!
            System.out.println(e.getMessage());
            e.printStackTrace();
        }
        return null;
    }

    @Override
    public String encryptToBase64(byte[] input, IvParameterSpec iv) {
        return Base64Wrap.encode(encryptToByteArray(input, iv));
    }

    @Override
    public String encryptToBase64(String input, IvParameterSpec iv) {
        try {
            return encryptToBase64(input.getBytes("UTF-8"), iv);
        } catch (UnsupportedEncodingException e) {
            e.printStackTrace();
        }
        return null;
    }

    @Override
    public byte[] decryptFromBase64(String input) {
        //return decryptFromByteArray(Base64.getDecoder().decode(input.getBytes("UTF-8")));
        byte[] data = Base64Wrap.decode(input);
        // first 16 bytes of data is iv
        byte[] iv_bytes = new byte[16];
        for (int i = 0; i < 16; i++) {
            iv_bytes[i] = data[i];
        }
        // rest is ciphertext
        byte[] ciphertext = new byte[data.length - 16];
        for (int i = 0; i < ciphertext.length; i++) {
            ciphertext[i] = data[i+16];
        }
        IvParameterSpec iv = new IvParameterSpec(iv_bytes);

        return decryptFromByteArray(ciphertext, iv);
    }

    @Override
    public String decryptFromBase64ToString(String input) {
        return new String(decryptFromBase64(input));
    }

    @Override
    public String decryptFromByteArrayToString(byte[] input, IvParameterSpec iv) {
        return new String(decryptFromByteArray(input, iv));
    }
}
