package net.black3r.bakalarka.bakalarka;

import android.app.Notification;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.app.Service;
import android.content.Context;
import android.content.Intent;
import android.hardware.Sensor;
import android.hardware.SensorEvent;
import android.hardware.SensorEventListener;
import android.hardware.SensorManager;
import android.support.v4.app.NotificationCompat;
import android.util.Base64;
import android.util.Log;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.net.Socket;
import java.util.HashMap;
import java.util.HashSet;

public class CommunicationThread extends Thread implements SensorEventListener {

    private Configuration configuration;
    private SensorEvent lastEvent;

    @Override
    public void onSensorChanged(SensorEvent event) {
//        Log.d("onSensorChanged", "Sensor event received!");
        lastEvent = event;
    }

    @Override
    public void onAccuracyChanged(Sensor sensor, int accuracy) {
        // we do not need to do anything here
    }

    class ALS {
        public HashSet<Integer> open_fds;
        public HashMap<Integer, Integer> real_fd;
        public int fd_count;

        ALS() throws UnsupportedEncodingException {
            this.open_fds = new HashSet<>();
            this.real_fd = new HashMap<>();
            this.fd_count = 47;
            encryption_active = false;
        }

        public boolean encryption_active;
        public Crypter crypter;
    }

    private boolean ended = false;

    public void shutdown() {
        try {
//            Log.d("clientShutdown", "Trying to shutdown the client!");
            clientSocket.close();
            ended = true;
        } catch (IOException e) {
//            Log.d("clientShutdown", "Caught exception!");
            e.printStackTrace();
        }
    }

    private Socket clientSocket;
    private BufferedReader input;
    private final String version = "BETA1";
    private PrintWriter output;
    private ALS als;
    private Context context;
    private String prefix = "/sdcard";
    public CommunicationThread(Context ctx, Socket clientSocket) throws UnsupportedEncodingException {
        this.context = ctx;
        this.clientSocket = clientSocket;
        this.configuration = new Configuration(ctx);
        als = new ALS();
        try {
//            Log.d("commConstructor", "Constructing communicator");
            this.input = new BufferedReader(new InputStreamReader(this.clientSocket.getInputStream()));
            this.output = new PrintWriter(this.clientSocket.getOutputStream(), true);
            this.clientSocket.getOutputStream();
        } catch (IOException e) {
//            Log.d("commConstructor", "Exception during constructing communicator");
            e.printStackTrace();
        }
    }

    public void write(String line) {
        if (!this.clientSocket.isClosed()) {
            this.output.write(line);
            this.output.flush();
        }
    }

    public void streamWriteLine(String line) {
        if (!this.clientSocket.isClosed()) {
            this.output.write(line);
            this.output.write('\n');
            this.output.flush();
        }
    }

    public void writeLine(String line) {
        if (this.als.encryption_active) {
//            Log.d("writeLine", "Encrypting: " + line);
            String encrypted = this.als.crypter.encryptToBase64(line);
//            Log.d("writeLine", "Encrypted: " + encrypted);
            this.streamWriteLine(encrypted);
        } else {
            this.streamWriteLine(line);
        }
    }

    public void writeLine(float number) {
        writeLine(Float.toString(number));
    }

    public void writeData(byte[] data) {
        if (this.als.encryption_active) {
            String encrypted = this.als.crypter.encryptToBase64(data);
            this.streamWriteLine(encrypted);
        } else {
            this.writeLine(Base64Wrap.encode(data));
        }
    }

    public void writeLine(int number) {
        writeLine(Integer.toString(number));
    }

    public void writeLine(long what) {
        writeLine(Long.toString(what));
    }

    private void handleGetAttr(String[] argv) {
//        Log.d("handleResult", "Getting attributes of : " + argv[1]);
        String fullPath = prefix + argv[1];
        long[] wat = JNILib.getattr(fullPath);
        for (int i = 0; i < wat.length; i++) {
            this.writeLine(wat[i]);
        }
//        Log.d("handleResult", "Successful!");
    }

    private void handleVersion(String[] argv) {
//        Log.d("handleResult", "Writing message: " + "BETA1");
        this.writeLine("BETA1");
//        Log.d("handleResult", "Successful!");
    }

    private void handleUnlink(String[] argv) {
//        Log.d("handleResult", "Unlinking file: " + argv[1]);
        String fullPath = prefix + argv[1];
        int result = JNILib.unlink(fullPath);
        this.writeLine(result);
    }

    private void handleRmdir(String[] argv) {
//        Log.d("handleResult", "Removing directory: " + argv[1]);
        String fullPath = prefix + argv[1];
        int result = JNILib.rmdir(fullPath);
        this.writeLine(result);
    }

    private void handle_result(String read) throws IOException {
//        Log.d("handleResult", "Received message: " + read);
        if (read == null) {
            return;
        }
        String[] argv = read.split(" ");
        int argc = argv.length;
        JNILib.error = argc;
        String cmd = argv[0];
        switch (cmd) {
            case "IDENTIFY":
                handleIdentify();
                break;
            case "EXTENSIONS":
                handleExtensions();
                break;
            case "VERSION":
                handleVersion(argv);
                break;
            case "PREFIX":
                handlePrefix(argv[1]);
                break;
            case "GETATTR":
                handleGetAttr(argv);
                break;
            case "UNLINK":
                handleUnlink(argv);
                break;
            case "RMDIR":
                handleRmdir(argv);
                break;
            case "RENAME":
                handleRename(argv);
                break;
            case "STATFS":
                handleStatFs(argv[1]);
                break;
            case "ACCESS":
                handleAccess(argv);
                break;
            case "UTIMENS":
                handleUtimens(argv);
                break;
            case "FALLOCATE":
                handleFallocate(argv);
                break;
            case "READDIR":
                handleReaddir(argv[1]);
                break;
            case "MKDIR":
                handleMkdir(argv);
                break;
            case "CANENCRYPT":
                handleCanEncrypt(argv[1]);
                break;
            case "ENCRYPT":
                handleEncrypt(argv);
                break;
            case "CRYPTHELLO":
                handleCrypthello(argv[1]);
                break;
            case "TRUNCATE":
                handleTruncate(argv);
                break;
            case "OPEN":
                handleOpen(argv);
                break;
            case "CREATE":
                handleCreate(argv);
                break;
            case "CLOSE":
                handleClose(argv[1]);
                break;
            case "READ":
                handleRead(argv);
                break;
            case "WRITE":
                handleWrite(argv);
                break;
            case "PAIR":
                handlePair(argv);
                break;
            case "STARTSENSOR":
                handleStartSensor();
                break;
            case "QUITSENSOR":
                handleQuitSensor();
                break;
            case "GETROTATION":
                handleGetRotation();
                break;
            default:
                handleUnknown();
                break;
        }
    }

    private void handleGetRotation() {
        float x = 0, y = 0, z = 0;
        if (lastEvent != null) {
            x = lastEvent.values[0];
            y = lastEvent.values[1];
            z = lastEvent.values[2];
        }
        this.writeLine(x);
        this.writeLine(y);
        this.writeLine(z);
    }

    private void handleQuitSensor() {
        SensorManager sm=(SensorManager)context.getSystemService(Context.SENSOR_SERVICE);
        sm.unregisterListener(this);
    }

    private void handleStartSensor() {
        SensorManager sm=(SensorManager)context.getSystemService(Context.SENSOR_SERVICE);
        Sensor rotSensor = sm.getDefaultSensor(Sensor.TYPE_ROTATION_VECTOR);
        sm.registerListener(this, rotSensor, SensorManager.SENSOR_DELAY_GAME);
    }

    private void handleExtensions() {
        this.writeLine(3);
        this.writeLine("AES-CBC-128-B64");
        this.writeLine("FUSE-FS-1");
        this.writeLine("ROTATIONSENSOR");
    }

    private void handlePrefix(String prfx) {
        File f = new File(prfx);
        if (f.isDirectory()) {
            this.prefix = prfx;
            this.writeLine("OK");
        } else {
            this.writeLine("FAIL");
        }
    }

    private void handleEncrypt(String[] argv) {
        String uuid = argv[1];
        String algorithm = argv[2];
        // we check if encryption is possible
        // we check if we support the requested algorithm
        boolean available = this.configuration.hasSecretFor(uuid) && algorithm.equals("AES-CBC-128-B64");
        if (available) {
            // we respond with OK
            writeLine("OK");
            writeLine(configuration.getUUID());
            als.encryption_active = true;
            als.crypter = new AESCrypter(this.configuration.getSecretBytesFor(uuid));
            // now we start the encryption and the rest of the communication is encrypted!
        } else {
            writeLine("FAIL");
            writeLine(configuration.getUUID());
            // and we don't start the encryption
        }
    }

    private void handleCanEncrypt(String uuid) {
        // first argument is the ID of the device
        if (this.configuration.hasSecretFor(uuid)) {
            writeLine("YES");
        } else {
            writeLine("NO");
        }
        writeLine(configuration.getUUID());
    }

    private void handlePair(String[] argv) {
        String target = argv[1];
        PendingIntent pendingIntent = PendingIntent.getActivity(this.context, 0, new Intent(this.context, MainActivity.class), 0);
        Notification notification = new NotificationCompat.Builder(this.context)
                .setSmallIcon(android.R.drawable.ic_menu_add)
                .setContentTitle("Linux communicator")
                .setContentText("A device with ID: " + target + " wants to pair with this device!")
                .setContentIntent(pendingIntent)
                .build();
        NotificationManager nm = (NotificationManager) this.context.getSystemService(Context.NOTIFICATION_SERVICE);
        nm.notify(4742, notification);
    }

    private void handleIdentify() {
        this.writeLine(this.configuration.getUUID());
    }

    private void handleUnknown() {
//        Log.d("handleResult", "Writing message: " + "UNKNOWN COMMAND");
        this.writeLine("UNKNOWN COMMAND");
//        Log.d("handleResult", "Successful!");
    }

    private void handleWrite(String[] argv) {
//        Log.d("handleResult", "Writing to file");
        Integer to = Integer.parseInt(argv[1]);
        Integer size = Integer.parseInt(argv[2]);
        Integer offset = Integer.parseInt(argv[3]);
        String buffer_base64 = argv[4];
        byte[] buffer = Base64Wrap.decode(buffer_base64);
//        Log.d("handleResult", "Decoded buffer: " + buffer);
        if (!als.open_fds.contains(to)) {
//            Log.d("handleResult", "Unknown file descriptor");
            this.writeLine(0);
        } else {
//            Log.d("handleResult", "Found file descriptor and it should be open");
            Integer fd = als.real_fd.get(to);
            int ret = JNILib.write(fd, buffer, size, offset);
            this.writeLine(ret);
        }
    }

    private void handleRead(String[] argv) {
//        Log.d("handleResult", "Reading file");
        Integer from = Integer.parseInt(argv[1]);
        Integer size = Integer.parseInt(argv[2]);
        Integer offset = Integer.parseInt(argv[3]);
//        Log.d("handleResult", "File descriptor is: " + from.toString());
//        Log.d("handleResult", "Required number of bytes read is: " + size.toString());
        if (!als.open_fds.contains(from)) {
//            Log.d("handleResult", "Unknown file descriptor");
            // TODO: possibly return some negative number as an error?
            this.writeLine(0);
            this.writeLine("");
        } else {
//            Log.d("handleResult", "Found file descriptor and it should be open");
            Integer fd = als.real_fd.get(from);
//            Log.d("handleResult", "Real file descriptor is: " + fd.toString());
            byte[] result = JNILib.read(fd, size, offset);
//            Log.d("handleResult", "JNI returned this errno: " + JNILib.error);
//            Log.d("handleResult", "JNI returned this string: " + result);
            String base64 = Base64.encodeToString(result, Base64.DEFAULT | Base64.NO_WRAP);
//            Log.d("handleResult", "BASE64 encoded: " + base64);
            int result_length = result.length;
            this.writeLine(result_length);
            this.writeLine(base64);
        }
    }

    private void handleClose(String string) {
//        Log.d("handleResult", "Closing file");
        Integer fd = Integer.parseInt(string);
        if (!als.open_fds.contains(fd)) {
//            Log.d("handleResult", "Unknown file descriptor");
            this.writeLine(-1);
        } else {
//            Log.d("handleResult", "Open file found, closing...");
            int rfd = als.real_fd.get(fd);
            int ret = JNILib.close(rfd);
            this.writeLine(ret);
            als.real_fd.remove(fd);
            als.open_fds.remove(fd);
        }
    }

    private void handleCreate(String[] argv) {
//        Log.d("handleResult", "Creating file");
        String fullPath = prefix + argv[1];
        Integer mode = Integer.parseInt(argv[2]);
        int fd = JNILib.create(fullPath, mode);
//        Log.d("handleResult", "Result is " + Integer.toString(fd));
        if (fd < 0) {
            this.writeLine(fd);
        } else {
            Integer new_fd = ++als.fd_count;
            als.real_fd.put(new_fd, fd);
            als.open_fds.add(new_fd);
//            Log.d("handleResult", "Returning " + new_fd.toString());
            this.writeLine(new_fd);
        }
    }

    private void handleOpen(String[] argv) {
//        Log.d("handleResult", "Opening file");
        String fullPath = prefix + argv[1];
        Integer args = Integer.parseInt(argv[2]);
        int fd = JNILib.open(fullPath, args.intValue());
//        Log.d("handleResult", "Result is " + Integer.toString(fd));
        if (fd < 0) {
            this.writeLine(fd);
        } else {
            Integer new_fd = ++als.fd_count;
            als.real_fd.put(new_fd, fd);
            als.open_fds.add(new_fd);
//            Log.d("handleResult", "Returning " + new_fd.toString());
            this.writeLine(new_fd);
        }
    }

    private void handleTruncate(String[] argv) {
//        Log.d("handleResult", "Truncating file");
        String fullPath = prefix + argv[1];
        Integer offset = Integer.parseInt(argv[2]);
        int ret = JNILib.truncate(fullPath, offset);
        this.writeLine(ret);
    }

    private void handleCrypthello(String s) {
        String incoming = s;
        this.writeLine(incoming);
    }

    private void handleMkdir(String[] argv) {
//        Log.d("handleResult", "Creating directory");
        String fullPath = prefix + argv[1];
        Integer mode = Integer.parseInt(argv[2]);
        int ret = JNILib.mkdir(fullPath, mode);
        this.writeLine(ret);
    }

    private void handleReaddir(String s) {
//        Log.d("handleResult", "Reading directory");
        String fullPath = prefix + s;
        File dir = new File(fullPath);
        String[] fileList = dir.list();
        this.writeLine(fileList.length);
        for (String f : fileList) {
//            Log.d("handleResult", f);
            this.writeLine(f);
        }
    }

    private void handleFallocate(String[] argv) {
//        Log.d("handleResult", "Calling fallocate on file " + argv[1]);
        Integer file = Integer.parseInt(argv[1]);
        Integer mode = Integer.parseInt(argv[2]);
        Integer offset = Integer.parseInt(argv[3]);
        Integer length = Integer.parseInt(argv[4]);
        if (!als.open_fds.contains(file)) {
//            Log.d("handleResult", "Unknown file descriptor");
            this.writeLine(9);
        } else {
//            Log.d("handleResult", "handling fallocate");
            int fd = als.real_fd.get(file);
            int result = JNILib.fallocate(fd, mode, offset, length);
//            Log.d("handleResult", "result is: " + Integer.toString(result));
            this.writeLine(result);
        }
    }

    private void handleUtimens(String[] argv) {
//        Log.d("handleResult", "Changing modification time of file: " + argv[1]);
        String fullPath = prefix + argv[1];
        Integer sec0 = Integer.parseInt(argv[2]);
        Integer nsec0 = Integer.parseInt(argv[3]);
        Integer sec1 = Integer.parseInt(argv[4]);
        Integer nsec1 = Integer.parseInt(argv[5]);
        int answer = JNILib.futimens(fullPath, sec0, nsec0, sec1, nsec1);
        this.writeLine(answer);
    }

    private void handleAccess(String[] argv) {
//        Log.d("handleResult", "Testing access of file" + argv[1]);
        String fullPath = prefix + argv[1];
        Integer flags = Integer.parseInt(argv[2]);
        int answer = JNILib.access(fullPath, flags);
        this.writeLine(answer);
    }

    private void handleStatFs(String s) {
//        Log.d("handleResult", "Getting attributes of file-system at: " + s);
        String fullPath = prefix + s;
        long[] wat = JNILib.statfs(fullPath);
        for (int i = 0; i < wat.length; i++) {
            this.writeLine(wat[i]);
        }
    }

    private void handleRename(String[] argv) {
//        Log.d("handleResult", "Renaming file " + argv[1] + " to " + argv[2]);
        String oldpath = prefix + argv[1];
        String newpath = prefix + argv[2];
        int result = JNILib.rename(oldpath, newpath);
        this.writeLine(result);
    }

    public String streamReadLine() {
        try {
            return this.input.readLine();
        } catch (IOException e) {
            e.printStackTrace();
        }
        return null;
    }

    public String readLine() {
        String x = streamReadLine();
        if (x == null) return null;
        if (this.als.encryption_active) {
//            Log.d("readLine", "Received encrypted line: " + x);
//            Log.d("readLine", "Decrypted: " + this.als.crypter.decryptFromBase64ToString(x));
            return this.als.crypter.decryptFromBase64ToString(x);
        } else {
            return x;
        }
    }

    public byte[] readData() {
        String x = streamReadLine();
        if (this.als.encryption_active) {
            return this.als.crypter.decryptFromBase64(x);
        } else {
            return Base64Wrap.decode(x);
        }
    }

    @Override
    public void run() {
        while(!Thread.currentThread().isInterrupted()) {
            try {
                if (ended) break;
                String read = this.readLine();
                if (read == null) continue;
//                Log.d("commRun", "Incoming message");
                handle_result(read);
            } catch (IOException e) {
//                Log.d("commRun", "Exception during incoming message");
                e.printStackTrace();
                break;
            }
        }
    }
}
