package net.black3r.bakalarka.bakalarka;

import android.content.Context;
import android.content.SharedPreferences;
import android.util.Log;
import android.widget.Toast;

import java.util.UUID;

/**
 * Created by roman on 29.5.2015.
 */
public class Configuration {
    private Context context;
    private String uuid;
    private SharedPreferences deviceInfo;
    private SharedPreferences secrets;

    public Configuration(Context ctx) {
        this.context = ctx;
        this.deviceInfo = context.getSharedPreferences("device-info", Context.MODE_PRIVATE);
        this.secrets = context.getSharedPreferences("device-secrets", Context.MODE_PRIVATE);
        // we make sure that if we are not set-up yet, we store the device-id immediately so it
        // doesn't change within the application lifetime.
        // device-id is used to identify the device in all clients configuration.
        if (!deviceInfo.contains("device-id")) {
            deviceInfo.edit().putString("device-id", java.util.UUID.randomUUID().toString()).commit();
        }
        this.uuid = deviceInfo.getString("device-id", java.util.UUID.randomUUID().toString());
    }

    public boolean hasSecretFor(String key) {
        return secrets.contains(key);
    }

    public String getSecretFor(String key) {
        return secrets.getString(key, "MDEyMzQ1Njc4OWFiY2RlZg==");
    }

    public byte[] getSecretBytesFor(String key) {
        String secret = getSecretFor(key);
//        Log.d("Configuration", "Base64 secret is: " + secret);
        return Base64Wrap.decode(secret);
    }

    public void storeSecretFor(String key, String secret) {
        secrets.edit().putString(key, secret).commit();
        Toast.makeText(context, "Successfully paired with a new device!", Toast.LENGTH_SHORT).show();
    }

    public String[] getSecretList() {
        return secrets.getAll().keySet().toArray(new String[0]);
    }

    public void removeSecretFor(String key) {
        secrets.edit().remove(key).commit();
    }

    public String getUUID() {
        return this.uuid;
    }
}
