package net.black3r.bakalarka.bakalarka;

import android.app.Notification;
import android.app.PendingIntent;
import android.app.Service;
import android.content.Context;
import android.content.Intent;
import android.net.DhcpInfo;
import android.net.wifi.WifiManager;
import android.os.IBinder;
import android.os.SystemClock;
import android.support.v4.app.NotificationCompat;
import android.util.Log;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.HashSet;
import java.util.Set;

public class ServerService extends Service {
    public static String STARTFOREGROUND_ACTION = "net.black3r.bakalarka.bakalarka.serverservice.action.start";
    public static String STOPFOREGROUND_ACTION = "net.black3r.bakalarka.bakalarka.serverservice.action.stop";

    private ServerSocket serverSocket;
    ServerThread serverThread = null;
    public static final int SERVERPORT = 6000;

    @Override
    public int onStartCommand(Intent intent, int flags, int startId) {
        if (intent.getAction().equals(STARTFOREGROUND_ACTION)) {

            Intent notificationIntent = new Intent(this, MainActivity.class);
            notificationIntent.setFlags(Intent.FLAG_ACTIVITY_NEW_TASK | Intent.FLAG_ACTIVITY_CLEAR_TASK);
            PendingIntent pendingIntent = PendingIntent.getActivity(this, 0, notificationIntent, 0);

            Intent closeIntent = new Intent(this, ServerService.class);
            closeIntent.setAction(STOPFOREGROUND_ACTION);
            PendingIntent pcloseIntent = PendingIntent.getService(this, 0, closeIntent, 0);

            Notification notification = new NotificationCompat.Builder(this)
                    .setSmallIcon(android.R.drawable.ic_dialog_info)
                    .setContentTitle("Linux communicator")
                    .setTicker("Linux communicator")
                    .setContentText("Linux communication server running")
                    .setContentIntent(pendingIntent)
                    .setOngoing(true)
                    .addAction(android.R.drawable.ic_menu_close_clear_cancel, "Stop", pcloseIntent)
                    .build();
            startForeground(4247, notification);
            this.serverThread = new ServerThread(this.getApplicationContext());
            this.serverThread.start();
        } else if(intent.getAction().equals(STOPFOREGROUND_ACTION)) {
            serverThread.shutdown();
            serverThread.interrupt();
            stopForeground(true);
            stopSelf();
        }
        return START_STICKY;
    }

    @Override
    public void onDestroy() {
        super.onDestroy();
    }

    class BroadcastThread extends Thread {

        private Context context;
        public int DISCOVERY_PORT = 9247;
        private DatagramSocket socket;
        private Configuration configuration;

        InetAddress getBroadcastAddress() throws IOException {
            WifiManager wifi = (WifiManager) context.getSystemService(Context.WIFI_SERVICE);
            DhcpInfo dhcp = wifi.getDhcpInfo();

            int broadcast = (dhcp.ipAddress & dhcp.netmask) | ~dhcp.netmask;
            byte[] quads = new byte[4];
            for (int k = 0; k < 4; k++)
                quads[k] = (byte) ((broadcast >> k * 8) & 0xFF);
            return InetAddress.getByAddress(quads);
        }

        public BroadcastThread(Context ctx) {
            this.context = ctx;
            this.configuration = new Configuration(context);
        }

        private boolean end = false;

        public void shutdown() {
            end = true;
            this.interrupt();
        }

        @Override
        public void run() {
            try {
//                Log.d("broadcastServer", "Creating server!");
                socket = new DatagramSocket();
                socket.setBroadcast(true);
            } catch (SocketException e) {
                e.printStackTrace();
            }

            while(!Thread.currentThread().isInterrupted()) {
                if (end) break;
//                Log.d("broadcastServer", "Sending a packet");
                try {
                    DatagramPacket packet = null;
                    String data = "ALC " + configuration.getUUID();
                    packet = new DatagramPacket(data.getBytes(), data.length(),
                            getBroadcastAddress(), DISCOVERY_PORT);
                    socket.send(packet);
                    Thread.sleep(1000);
                } catch (IOException e) {
                    e.printStackTrace();
//                    Log.d("broadcastServer", "IO Exception caught! Quitting...");
                    break;
                } catch (InterruptedException e) {
                    e.printStackTrace();
//                    Log.d("broadcastServer", "IO Exception caught! Quitting...");
                    break;
                }
            }
        }
    }

    class ServerThread extends Thread {

        Context context;
        private Set<CommunicationThread> connections;
        private BroadcastThread broadcastThread;

        public ServerThread(Context applicationContext) {
            this.context = applicationContext;
        }
        private Socket socket;

        public void shutdown() {
            try {
//                Log.d("serverShutdown", "Trying to shutdown the server!");
                serverSocket.close();
//                Log.d("serverShutdown", "Trying to shutdown all connections!");
                for (CommunicationThread client : connections) {
                    // TODO: shutdown connection gracefully (after finishing a command)
                    client.shutdown();
                }
//                Log.d("serverShutdown", "Trying to shutdown broadcast server");
                broadcastThread.shutdown();
            } catch (IOException e) {
//                Log.d("serverShutdown", "Caught exception!");
                e.printStackTrace();
            }
        }

        @Override
        public void run() {
            socket = null;
            connections = new HashSet<>();
            broadcastThread = new BroadcastThread(context);
            broadcastThread.start();
            try {
//                Log.d("serverRun", "Creating server");
                serverSocket = new ServerSocket(SERVERPORT);
            } catch (IOException e) {
//                Log.d("serverRun", "Exception during server creation");
                e.printStackTrace();
            }
            while (!Thread.currentThread().isInterrupted()) {
                try {
                    if (serverSocket.isClosed()) {
                        break;
                    }
//                    Log.d("serverRun", "Incoming connection!");
                    socket = serverSocket.accept();
                    CommunicationThread commThread = new CommunicationThread(context, socket);
                    connections.add(commThread);
                    commThread.start();
                } catch (IOException e) {
//                    Log.d("serverRun", "Exception during incoming connection");
                    e.printStackTrace();
                }
            }
//            Log.d("serverRun", "stopping server!");
            try {
                // try to close socket when thread is interrupted
//                Log.d("serverRun", "closing the socket!");
                serverSocket.close();
            } catch (IOException e) {
//                Log.d("serverRun", "exception caught!");
                e.printStackTrace();
            }
        }
    }

    public ServerService() {

    }

    @Override
    public IBinder onBind(Intent intent) {
        throw new UnsupportedOperationException("Not yet implemented");
    }


}
