/*
 * Copyright (C) 2009 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// OpenGL ES 2.0 code

#include <jni.h>
#include <android/log.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/statvfs.h>
#include <unistd.h>
using namespace std;

#define  LOG_TAG    "libbakalarka"
#define  LOGI(...)  __android_log_print(ANDROID_LOG_INFO,LOG_TAG,__VA_ARGS__)
#define  LOGE(...)  __android_log_print(ANDROID_LOG_ERROR,LOG_TAG,__VA_ARGS__)

extern "C" {
    JNIEXPORT void JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_createFile(JNIEnv *env, jobject obj);
    JNIEXPORT jlongArray JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_getattr(JNIEnv *env, jobject obj, jstring path);
    JNIEXPORT jlongArray JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_statfs(JNIEnv *env, jobject obj, jstring path);
    JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_open(JNIEnv *env, jobject obj, jstring path, jint flags);
    JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_close(JNIEnv *env, jobject obj, jint fd);
    JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_create(JNIEnv *env, jobject obj, jstring path, jint mode);
    JNIEXPORT jbyteArray JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_read(JNIEnv *env, jobject obj, jint fd, jint size, jint offset);
    JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_write(JNIEnv *env, jobject obj, jint fd, jbyteArray buf, jint size, jint offset);
    JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_mkdir(JNIEnv *env, jobject obj, jstring path, jint mode);
    JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_truncate(JNIEnv *env, jobject obj, jstring path, jint offset);
    JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_unlink(JNIEnv *env, jobject obj, jstring path);
    JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_rmdir(JNIEnv *env, jobject obj, jstring path);
    JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_rename(JNIEnv *env, jobject obj, jstring oldpath, jstring newpath);
    JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_access(JNIEnv *env, jobject obj, jstring path, jint flags);
    JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_fallocate(JNIEnv *env, jobject obj, jint fd, jint mode, jint offset, jint length);
    JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_futimens(JNIEnv *env, jobject obj, jstring path, jint sec0, jint nsec0, jint sec1, jint nsec1);
}

JNIEXPORT void JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_createFile(JNIEnv *env, jobject obj)
{
//    LOGI("Hello world!");
    FILE* f = fopen("/sdcard/test.txt", "w+");
//    LOGI("Return: %d", f);
    fwrite("hello world\n", strlen("hello world\n"), 1, f);
    if (f != 0) {
        fclose(f);
    } else {
//        LOGI("Error: %d", errno);
    }
//    LOGI("DONE!");
}

jbyteArray as_byte_array(JNIEnv *env, unsigned char* buf, int len) {
    jbyteArray array = env->NewByteArray (len);
    env->SetByteArrayRegion (array, 0, len, reinterpret_cast<jbyte*>(buf));
    return array;
}

jcharArray as_char_array(JNIEnv *env, char* buf, int len) {
    jcharArray array = env->NewCharArray(len);
    env->SetCharArrayRegion(array, 0, len, reinterpret_cast<jchar*>(buf));
    return array;
}

jlongArray as_long_array(JNIEnv *env, long long* buf, int len) {
    jlongArray array = env->NewLongArray(len);
    env->SetLongArrayRegion(array, 0, len, reinterpret_cast<jlong*>(buf));
    return array;
}

unsigned char* as_unsigned_char_array(JNIEnv *env, jbyteArray array) {
    int len = env->GetArrayLength (array);
    // TODO: free this
    unsigned char* buf = new unsigned char[len];
    env->GetByteArrayRegion (array, 0, len, reinterpret_cast<jbyte*>(buf));
    return buf;
}

JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_open(JNIEnv *env, jobject obj, jstring path, jint flags) {
    const char *c_path;
    c_path = env->GetStringUTFChars(path, NULL);
    int f = open(c_path, flags);
    delete c_path;
    return f;
}

JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_access(JNIEnv *env, jobject obj, jstring path, jint flags) {
    const char *c_path;
    c_path = env->GetStringUTFChars(path, NULL);
    int f = access(c_path, flags);
    delete c_path;
    if (f != 0) return errno;
    return f;
}

// ok i created this based on the official posix documentation of create, hope that's okay now!
JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_create(JNIEnv *env, jobject obj, jstring path, jint mode) {
    const char *c_path;
    c_path = env->GetStringUTFChars(path, NULL);
    int f = open(c_path, O_WRONLY|O_CREAT|O_TRUNC, mode);
    delete c_path;
    return f;
}

JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_mkdir(JNIEnv *env, jobject obj, jstring path, jint mode) {
    const char *c_path;
    c_path = env->GetStringUTFChars(path, NULL);
    int f = mkdir(c_path, mode);
    delete c_path;
    if (f < 0) return errno;
    return f;
}

JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_truncate(JNIEnv *env, jobject obj, jstring path, jint offset) {
    const char *c_path;
    c_path = env->GetStringUTFChars(path, NULL);
    int f = truncate(c_path, offset);
    delete c_path;
    if (f < 0) return errno;
    return f;
}

JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_rename(JNIEnv *env, jobject obj, jstring oldpath, jstring newpath) {
    const char *c_path_old;
    c_path_old = env->GetStringUTFChars(oldpath, NULL);
    const char *c_path_new;
    c_path_new = env->GetStringUTFChars(newpath, NULL);
    int f = rename(c_path_old, c_path_new);
    delete c_path_old;
    delete c_path_new;
    if (f < 0) return errno;
    return f;
}

JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_close(JNIEnv *env, jobject obj, jint fd) {
    return close(fd);
}

JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_unlink(JNIEnv *env, jobject obj, jstring path) {
    const char *c_path;
    c_path = env->GetStringUTFChars(path, NULL);
    int f = unlink(c_path);
    delete c_path;
    if (f < 0) return errno;
    return f;
}

JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_rmdir(JNIEnv *env, jobject obj, jstring path) {
    const char *c_path;
    c_path = env->GetStringUTFChars(path, NULL);
    int f = rmdir(c_path);
    delete c_path;
    if (f != 0) return errno;
    return f;
}

JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_fallocate(JNIEnv *env, jobject obj, jint fd, jint mode, jint offset, jint length) {
    int result = fallocate(fd, mode, offset, length);
    if (result != 0) return errno;
    return 0;
}

JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_futimens(JNIEnv *env, jobject obj, jstring path, jint sec0, jint nsec0, jint sec1, jint nsec1) {
    const char *c_path;
    c_path = env->GetStringUTFChars(path, NULL);
    int fd = open(c_path, O_RDWR);
    if (fd < 0) {
        return errno;
    }
    struct timespec tv[2];
    tv[0].tv_sec = sec0;
    tv[0].tv_nsec = nsec0;
    tv[1].tv_sec = sec1;
    tv[1].tv_nsec = nsec1;
    int ret = futimens(fd, tv);
    close(fd);
//    LOGI("FUTIMENS RESULT: %d\n", errno);
    if (ret != 0) return errno;
    return 0;
}

JNIEXPORT jbyteArray JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_read(JNIEnv *env, jobject obj, jint fd, jint size, jint offset) {
//    LOGI("About to read %d bytes at offset: %d\n", size, offset);
    int f = fd;
    unsigned char buffer[size];
    int read_bytes = pread(fd, buffer, size, offset);
    jclass cls = env->FindClass("net/black3r/bakalarka/bakalarka/JNILib");
    if (cls == NULL) {
//        LOGI("UNABLE TO LOAD CLASS");
    }
    jfieldID jFieldId = env->GetStaticFieldID(cls,"error","I");
    if(jFieldId == 0) {
//        LOGI("UNABLE TO ACCESS ERRNO FIELD");
    } else {
        env->SetStaticIntField(cls,jFieldId, read_bytes < 0 ? errno : 0);
    }
    if (read_bytes == -1) {
        LOGI("READ ERROR: %d\n", errno);
        return as_byte_array(env, (unsigned char*) buffer, 0);
    } else {
//        LOGI("Read %d bytes: %s \n", read_bytes, buffer);
        return as_byte_array(env, (unsigned char*) buffer, read_bytes);
        //char buffer2[read_bytes];
        //strncpy(buffer2, buffer, read_bytes+1);
        //buffer2[read_bytes] = 0;
    }
}

JNIEXPORT jint JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_write(JNIEnv *env, jobject obj, jint fd, jbyteArray buf, jint size, jint offset) {
    int f = fd;
    unsigned char *c_buf = as_unsigned_char_array(env, buf);
    //c_buf = env->GetStringUTFChars(buf, NULL);
    int ret = pwrite(fd, c_buf, size, offset);
    delete [] c_buf;
    return ret;
}

JNIEXPORT jlongArray JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_getattr(JNIEnv *env, jobject obj, jstring path) {
    const char *c_path;

    c_path = env->GetStringUTFChars(path, NULL);
    struct stat *fileStatp = new struct stat;
    int retc = stat(c_path, fileStatp);
    delete c_path;

    if (retc < 0) {
//        LOGI("fstat failed!");
    }
    int size = 14;
    long long data[] = {
        retc == 0 ? 0 : errno,
        fileStatp->st_dev,
        fileStatp->st_ino,
        fileStatp->st_mode,
        fileStatp->st_nlink,
        fileStatp->st_uid,
        fileStatp->st_gid,
        fileStatp->st_rdev,
        fileStatp->st_size,
        fileStatp->st_blksize,
        fileStatp->st_blocks,
        fileStatp->st_atime,
        fileStatp->st_mtime,
        fileStatp->st_ctime,
    };

    for (int i = 0; i < size; i++) {
//        LOGI("Data: %lld", data[i]);
    }

    delete fileStatp;

    return as_long_array(env, data, size);
}

JNIEXPORT jlongArray JNICALL Java_net_black3r_bakalarka_bakalarka_JNILib_statfs(JNIEnv *env, jobject obj, jstring path) {
    const char *c_path;

    c_path = env->GetStringUTFChars(path, NULL);
    struct statvfs *fileStatp = new struct statvfs;
    int retc = statvfs(c_path, fileStatp);
    delete c_path;

    if (retc < 0) {
//        LOGI("statvfs failed!");
    }
    int size = 12;
    long long data[] = {
        retc == 0 ? 0 : errno,
        fileStatp->f_bsize,
        fileStatp->f_frsize,
        fileStatp->f_blocks,
        fileStatp->f_bfree,
        fileStatp->f_bavail,
        fileStatp->f_files,
        fileStatp->f_ffree,
        fileStatp->f_favail,
        fileStatp->f_fsid,
        fileStatp->f_flag,
        fileStatp->f_namemax,
    };

    for (int i = 0; i < size; i++) {
//        LOGI("Data: %lld", data[i]);
    }

    delete fileStatp;
    return as_long_array(env, data, size);
}