#include <openssl/conf.h>
#include <openssl/evp.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#include <cstring>

#include "aes_crypto.h"
#include "base64.h"
#include "alc_constants.h"
#include "libalc.h"

using namespace std;

int generate_iv(unsigned char* iv) {
    return RAND_bytes(iv, 16);
}

int encrypt(unsigned char *plaintext, size_t plaintext_len, unsigned char *key, unsigned char *iv,
            unsigned char *ciphertext) {
    EVP_CIPHER_CTX *ctx;
    int len;
    int ciphertext_len;

    if (!(ctx = EVP_CIPHER_CTX_new())) handle_crypto_errors();
    if (1 != EVP_EncryptInit_ex(ctx, EVP_aes_128_cbc(), NULL, key, iv))
        handle_crypto_errors();

    if (1 != EVP_EncryptUpdate(ctx, ciphertext, &len, plaintext, plaintext_len))
        handle_crypto_errors();

    ciphertext_len = len;
    if (1 != EVP_EncryptFinal_ex(ctx, ciphertext + len, &len)) handle_crypto_errors();
    ciphertext_len += len;

    EVP_CIPHER_CTX_free(ctx);

    return ciphertext_len;
}

int decrypt(unsigned char *ciphertext, int ciphertext_len, unsigned char *key, unsigned char *iv, unsigned char *plaintext) {
    EVP_CIPHER_CTX *ctx;

    int len;

    int plaintext_len;

    if(!(ctx = EVP_CIPHER_CTX_new())) handle_crypto_errors();

    if(1 != EVP_DecryptInit_ex(ctx, EVP_aes_128_cbc(), NULL, key, iv))
        handle_crypto_errors();

    if(1 != EVP_DecryptUpdate(ctx, plaintext, &len, ciphertext, ciphertext_len))
        handle_crypto_errors();

    plaintext_len = len;

    if(1 != EVP_DecryptFinal_ex(ctx, plaintext + len, &len)) handle_crypto_errors();
    plaintext_len += len;

    EVP_CIPHER_CTX_free(ctx);

    return plaintext_len;
}

unsigned char* global_b64_buffer = NULL;

std::string encrypt_base64(unsigned char *plaintext, size_t plaintext_len, unsigned char *key, unsigned char *iv) {
    if (global_b64_buffer == NULL) {
        global_b64_buffer = (unsigned char *) malloc(MAX_COMMAND_LENGTH);
    }
    unsigned char *buffer = global_b64_buffer;
    int ciphertext_len = encrypt(plaintext, plaintext_len, key, iv, buffer+16);
    memcpy(buffer, iv, 16);
    return base64_encode(buffer, ciphertext_len + 16);
}

std::string encrypt_base64(unsigned char *plaintext, size_t plaintext_len, unsigned char *key) {
    unsigned char iv_bytes[16];
    generate_iv(iv_bytes);
    return encrypt_base64(plaintext, plaintext_len, key, iv_bytes);
}

int decrypt_base64(std::string ciphertext, unsigned char* key, unsigned char* plaintext) {
    string decoded = base64_decode(ciphertext);
    string iv_str = decoded.substr(0, 16);
    string cipher_str = decoded.substr(16, decoded.length() - 16);
    return decrypt((unsigned char *) cipher_str.c_str(), cipher_str.length(), key, (unsigned char *) iv_str.c_str(), plaintext);
}

int decrypt_base64(unsigned char *ciphertext, int ciphertext_len, unsigned char* key, unsigned char* plaintext) {
    std::string str((const char*) ciphertext, ciphertext_len);
    return decrypt_base64(str, key, plaintext);
}

void init_crypto() {
    ERR_load_crypto_strings();
    OpenSSL_add_all_algorithms();
    OPENSSL_config(NULL);
}

void destroy_crypto() {
    EVP_cleanup();
    ERR_free_strings();
}

void handle_crypto_errors(void) {
    ERR_print_errors_fp(stderr);
    fatal_error("Cryptographic error. Probably using wrong encryption key on either side of communication. Re-pair needed!");
}
