#include "alc_config.h"
#include "mkpath.h"
#include "uuid.h"
#include "alc_log.h"
#include <fstream>

using namespace std;

string alc_uuid() {
    string homedir(getenv("HOME"));
    string alc_config_basedir = homedir + "/.config/alc/";
//    LOGI("obtained basedir\n");
    int ret;
    if ((ret = mkpath(alc_config_basedir, 0774)) != 0 && errno != EEXIST) {
//        LOGI("created folder failed!: %d %d\n", ret, errno);
        return "";
    } else {
//        LOGI("created folder successfullly!\n");
        string uuid_conf_path = alc_config_basedir + "uuid.conf";
        ifstream uuid_conf_file(uuid_conf_path);
        if (!uuid_conf_file.is_open()) {
//            LOGI("file is not open\n");
            // we create file, generate random UUID and write it
            ofstream uuid_conf_file_write(uuid_conf_path);
            string uuid = generate_uuid_string();
//            LOGI("Generated UUID: %s\n", uuid.c_str());
            uuid_conf_file_write << uuid;
            uuid_conf_file_write.close();
            return uuid;
        } else {
//            LOGI("file is open\n");
            // we read the UUID from the file
            string uuid;
            uuid_conf_file >> uuid;
            uuid_conf_file.close();
            return uuid;
        }
    }
}

string get_keystore_path() {
    string homedir(getenv("HOME"));
    return homedir + "/.config/alc/keys.conf";
}

/**
 * returns true if successful, false otherwise
 */
bool create_keystore_if_not_exists() {
    string homedir(getenv("HOME"));
    string alc_config_basedir = homedir + "/.config/alc/";
//    LOGI("obtained basedir\n");
    int ret;
    if ((ret = mkpath(alc_config_basedir, 0774)) != 0 && errno != EEXIST) {
//        LOGI("created folder failed!: %d %d\n", ret, errno);
        return false;
    } else {
//        LOGI("created folder successfullly!\n");
        string key_conf_path = alc_config_basedir + "keys.conf";
        ifstream key_conf_file(key_conf_path);
        if (!key_conf_file.is_open()) {
//            LOGI("file is not open\n");
            // we create file, generate random UUID and write it
            ofstream key_conf_file_write(key_conf_path);
            key_conf_file_write.close();
            return true;
        } else {
//            LOGI("file is open\n");
            // we read the UUID from the file
            key_conf_file.close();
            return true;
        }
    }
}

map<string, string> get_keystore() {
    create_keystore_if_not_exists();
    string path = get_keystore_path();
//    LOGI("Path is: %s\n", path.c_str());
    ifstream key_conf_file(path);
    if (!key_conf_file.is_open()) {
//        LOGI("Unable to open keystore file\n");
        return map<string, string>();
    } else {
        map<string, string> result;
        string uuid, key;
        while(key_conf_file >> uuid >> key) {
//            LOGI("Reading UUID: %s\n", uuid.c_str());
//            LOGI("Reading Key: %s\n", key.c_str());
            result[uuid] = key;
        }
        key_conf_file.close();
        return result;
    }
}

string get_key(string uuid) {
    map<string, string> keystore = get_keystore();
    return keystore[uuid];
}

bool has_key_for(string uuid) {
    map<string, string> keystore = get_keystore();
    return keystore.find(uuid) != keystore.end();
}

bool write_keystore(map<string, string> keystore) {
    create_keystore_if_not_exists();
    string path = get_keystore_path();
    ofstream key_conf_file(path);
    if (!key_conf_file.is_open()) {
        return false;
    }
    for (pair<string, string> keypair : keystore) {
        key_conf_file << keypair.first << " " << keypair.second << endl;
    }
    key_conf_file.close();
    return true;
}

bool add_keypair(string uuid, string key) {
    map<string, string> keystore = get_keystore();
    keystore[uuid] = key;
    return write_keystore(keystore);
}