#include <string>
#include <vector>
#include "fuse_impl.h"
#include "libalc.h"
#include "alc_constants.h"

int afs_open_fuse(const char *path, struct fuse_file_info *fi) {
    LLS* lls = (LLS *) fuse_get_context()->private_data;
    int ret = afs_open_fd(path, fi->flags, lls);
    if (ret < 0) return ret;
    fi->fh = ret;
    return 0;
}

int afs_mkdir_fuse(const char *path, mode_t mode) {
    LLS* lls = (LLS *) fuse_get_context()->private_data;
    return afs_mkdir(path, mode, lls);
}

int afs_truncate_fuse(const char *path, off_t offset) {
    LLS* lls = (LLS *) fuse_get_context()->private_data;
    return afs_truncate(path, offset, lls);
}

int afs_release_fuse(const char *path, struct fuse_file_info *fi) {
    LLS* lls = (LLS *) fuse_get_context()->private_data;
    return afs_close_fd(fi->fh, lls);
}

int afs_fallocate_fuse(const char *path, int mode, off_t offset, off_t len, struct fuse_file_info *fi) {
    LLS* lls = (LLS *) fuse_get_context()->private_data;
    int fd = fi->fh;
    return afs_fallocate_fd(fd, mode, offset, len, lls);
}

int afs_read_fuse(const char *path, char *buf, size_t size, off_t offset, struct fuse_file_info *fi) {
    int fd = fi->fh;
    LLS* lls = (LLS *) fuse_get_context()->private_data;
    return afs_read_fd(fd, size, offset, buf, lls);
}

int afs_write_fuse(const char *path, const char *buf, size_t size, off_t offset, struct fuse_file_info *fi) {
    int fd = fi->fh;
    LLS* lls = (LLS *) fuse_get_context()->private_data;
    return afs_write_fd(fd, buf, size, offset, lls);
}

int afs_create_fuse(const char *path, mode_t mode, struct fuse_file_info *fi) {
    LLS* lls = (LLS *) fuse_get_context()->private_data;
    int ret = afs_create_fd(path, mode, lls);
    if (ret < 0) return ret;
    fi->fh = ret;
    return 0;
}

int afs_unlink_fuse(const char *path) {
    LLS* lls = (LLS *) fuse_get_context()->private_data;
    return afs_unlink(path, lls);
}

int afs_rmdir_fuse(const char *path) {
    LLS* lls = (LLS *) fuse_get_context()->private_data;
    return afs_rmdir(path, lls);
}

int afs_access_fuse(const char *path, int flags) {
    LLS* lls = (LLS *) fuse_get_context()->private_data;
    return afs_access(path, flags, lls);
}

int afs_utimens_fuse(const char *path, const struct timespec tv[2]) {
    LLS* lls = (LLS *) fuse_get_context()->private_data;
    return afs_futimens(path, tv, lls);
}

int afs_getattr_fuse(const char *path, struct stat *stbuf) {
    LLS* lls = (LLS *) fuse_get_context()->private_data;
    return afs_getattr(path, stbuf, lls);
}

int afs_statfs_fuse(const char *path, struct statvfs *stbuf) {
    LLS* lls = (LLS *) fuse_get_context()->private_data;
    return afs_statfs(path, stbuf, lls);
}

int afs_rename_fuse(const char *oldpath, const char *newpath) {
    LLS* lls = (LLS *) fuse_get_context()->private_data;
    return afs_rename(oldpath, newpath, lls);
}

int afs_readdir_fuse(const char *path, void *buf, fuse_fill_dir_t filler,
                            off_t offset, struct fuse_file_info *fi) {
    LLS* lls = (LLS *) fuse_get_context()->private_data;
    (void) offset;
    (void) fi;
    filler(buf, ".", NULL, 0);
    filler(buf, "..", NULL, 0);
    send_command(lls, {READDIR_COMMAND, path});
    int n = read_command_integer(lls);
    while(n--) {
        std::string x = read_command_string(lls);
        //LOG("Received filename: %s\n", x.c_str());
        filler(buf, x.c_str(), NULL, 0);
    }
    return 0;
}

struct fuse_operations afs_oper = {
        .getattr = afs_getattr_fuse,
        .readdir = afs_readdir_fuse,
        .open = afs_open_fuse,
        .read = afs_read_fuse,
        .release = afs_release_fuse,
        .write = afs_write_fuse,
        .create = afs_create_fuse,
        .mkdir = afs_mkdir_fuse,
        .truncate = afs_truncate_fuse,
        .statfs = afs_statfs_fuse,
        .unlink = afs_unlink_fuse,
        .rmdir = afs_rmdir_fuse,
        .rename = afs_rename_fuse,
        .access = afs_access_fuse,
        .fallocate = afs_fallocate_fuse,
        .utimens = afs_utimens_fuse,
};