#ifndef ANDROIDFUSE_LIBALC_H
#define ANDROIDFUSE_LIBALC_H

#include <string>
#include <vector>
#include <netinet/in.h>
#include "alc_log.h"

// TODO: vyfaktorovat UI string-y do nejakeho strings.h suboru
#define READ_COMMAND_FAILED_MSG "Read command failed"

struct LLS {
    bool encrypted = true;
    std::string *protocol_version;
    int remote_port;
    sockaddr_in remote_ip_addr;
    int comm_socket;
    std::string encryption_key_string =  "0123456789abcdef";
    unsigned char encryption_key_bytes[16];
    std::string encryption_type = "AES";
};

LLS * alc_connect(const char *ip_address, int port);
void alc_disconnect(const LLS *lls);
void alc_pair(LLS *lls);
bool alc_canencrypt(LLS* lls);
bool alc_startencrypt(LLS* lls);
void alc_toggleencrypt(LLS *lls, std::string uuid);
bool alc_crypthello(LLS *lls);
std::vector<std::string> alc_get_supported_extensions(LLS *lls);

void alc_enablesensors(LLS *lls);
void alc_disablesensors(LLS *lls);
std::vector<double> alc_getrotation(LLS *lls);

int net_read(LLS* config, char* buffer, size_t num_bytes);
int net_write(LLS* config, char* buffer, size_t num_bytes);
int net_write(LLS* config, const char* message);
int net_write(LLS* config, std::string buffer);
int send_command(LLS* config, std::vector<std::string> commands);
void read_command(LLS* config, std::vector<std::string> *target);
std::vector <std::string> * read_command(LLS* config);
std::string read_command_string(LLS* config);
int read_command_integer(LLS* config);
long read_command_long(LLS* config);
double read_command_double(LLS* config);
long long read_command_long_long(LLS* config);

void fatal_error(const char *msg);

int afs_open_fd(const char *path, int flags, LLS* lls);
int afs_close_fd(int fd, LLS* lls);
int afs_mkdir(const char *path, mode_t mode, LLS *lls);
int afs_truncate(const char *path, off_t offset, LLS *lls);
int afs_fallocate_fd(int fd, int mode, off_t offset, off_t len, LLS* lls);
int afs_read_fd(int fd, size_t size, off_t offset, char* buf, LLS* lls);
int afs_write_fd(int fd, const char* buf, size_t size, off_t offset, LLS* lls);
int afs_create_fd(const char *path, mode_t mode, LLS* lls);
int afs_unlink(const char *path, LLS *lls);
int afs_rmdir(const char *path, LLS *lls);
int afs_access(const char *path, int flags, LLS *lls);
int afs_futimens(const char *path, const struct timespec times[2], LLS* lls);
int afs_statfs(const char *path, struct statvfs *stbuf, LLS *lls);
int afs_getattr(const char *path, struct stat *stbuf, LLS *lls);
int afs_rename(const char *oldpath, const char *newpath, LLS *lls);
std::string afs_identify_phone(LLS* lls);
bool afs_changeroot(LLS *lls, std::string newroot);


#endif //ANDROIDFUSE_LIBALC_H
