#define FUSE_USE_VERSION 30
#include <fuse.h>
#include <stdio.h>
#include <string.h>
#include <string>
#include <vector>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>

#include "alc_constants.h"
#include "libalc.h"
#include "fuse_impl.h"

using namespace std;

int main(int argc, char *argv[])
{
    char* ipvalue = NULL;
    int portvalue = 6000;
    char* remote_dir = "/sdcard";
    char* local_dir = NULL;
    int c;
    opterr = 0;
    while ((c = getopt(argc, argv, "i:p:r:t:")) != -1) {
        switch (c) {
            case 'i':
                ipvalue = optarg;
                break;
            case 'p':
                portvalue = atoi(optarg);
                break;
            case 'r':
                remote_dir = optarg;
                break;
            case 't':
                local_dir = optarg;
                break;
            case '?':
                fatal_error("Unknown argument!");
                break;
            default:
                exit(1);
        }
    }
    if (ipvalue == NULL) {
        fprintf(stderr, "IP address is required!\n");
        exit(1);
    }
    if (local_dir == NULL) {
        fprintf(stderr, "Mountpoint is required!\n");
        exit(1);
    }

    int n;
    LLS* lls = alc_connect(ipvalue, portvalue);
    if (lls == NULL) {
        fatal_error("Unable to establish connection!\n");
    }

    if ((n = send_command(lls, {VERSION_COMMAND})) < 0) {
        perror("Unable to send command");
        fatal_error("Unable to send command");
    }

    string result = read_command_string(lls);
    if (!alc_canencrypt(lls)) {
        fatal_error("No encryption key found for this device. Please pair first\n");
    }
    if (!alc_startencrypt(lls)) {
        fatal_error("Unable to start encryption");
    };
    if (!alc_crypthello(lls)) {
        fatal_error("Encryption error!");
    }
    if (!afs_changeroot(lls, string(remote_dir))) {
        fprintf(stderr, "Unable to set root to: %s! Using /sdcard\n", remote_dir);
    }

    char option[] = "-s";
    char* arguments[] = {option, local_dir};

    return fuse_main(2, arguments, &afs_oper, lls);
}