#include <netdb.h>
#include <string.h>
#include <string>
#include <vector>
#include <unistd.h>
#include <sys/stat.h>
#include "libalc.h"
#include "alc_constants.h"
#include "alc_config.h"

using namespace std;

int main(int argc, char** argv) {
    int n;
    LLS* lls = alc_connect("192.168.47.100", 6000);

    if (lls == NULL) {
        fatal_error("Unable to establish connection!\n");
    }

    if ((n = send_command(lls, {VERSION_COMMAND})) < 0) {
        perror("Unable to send command");
    } else {
        LOGI("Successfully sent %d bytes!\n", n);
    }
    string result = read_command_string(lls);
    LOGI("Received message: %s\n", result.c_str());

    LOGI("Trying readdir!\n");
    send_command(lls, {READDIR_COMMAND, "/"});
    n = read_command_integer(lls);
    LOGI("File count: %d\n", n);
    while(n--) {
        string x = read_command_string(lls);
        LOGI("Received filename: %s\n", x.c_str());
    }

    LOGI("Trying getattr\n");

    struct stat st;
    afs_getattr("/test.txt", &st, lls);
    afs_getattr("/", &st, lls);
    afs_getattr("/neexistuje", &st, lls);

    LOGI("Trying open\n");

    int ret;
    int fd = afs_open_fd("/test.cc", 32768, lls);
    LOGI("OPEN returned: %d\n", fd);
    LOGI("Trying read\n");
    char buffer[4096];
    ret = afs_read_fd(fd, 4096, 0, buffer, lls);
    LOGI("Read %d bytes\n", ret);
    LOGI("Received string: %s\n", buffer);
    LOGI("everything is ok?\n");

    LOGI("Trying to read empty fajl");
    fd = afs_open_fd("/wat", 32768, lls);
    LOGI("OPEN returned: %d\n", fd);
    LOGI("Trying read\n");
    ret = afs_read_fd(fd, 4096, 0, buffer, lls);
    LOGI("Read %d bytes\n", ret);
    LOGI("Received string: %s\n", buffer);
    LOGI("everything is ok?\n");

    LOGI("Trying identify: \n");
    string x = afs_identify_phone(lls);
    LOGI("Received: %s\n", x.c_str());

    LOGI("Trying my uuid: %s\n", alc_uuid().c_str());
    LOGI("Trying canencrypt: %d\n", alc_canencrypt(lls));
    LOGI("Trying encrypt: %d\n", alc_startencrypt(lls));

    LOGI("Trying crypthello: %d\n", alc_crypthello(lls));

    LOGI("Trying change prefix: %d\n", afs_changeroot(lls, "/sdcard1"));
    LOGI("Trying readdir of the new root: \n");

    send_command(lls, {READDIR_COMMAND, "/"});
    n = read_command_integer(lls);
    LOGI("File count: %d\n", n);
    while(n--) {
        string x = read_command_string(lls);
        LOGI("Received filename: %s\n", x.c_str());
    }

    LOGI("Getting extension list: %d\n");
    vector<string> extlist = alc_get_supported_extensions(lls);
    for (string ext: extlist) {
        LOGI(ext.c_str());
        LOGI("\n");
    }

    LOGI("Starting sensors: \n");
    alc_enablesensors(lls);
    for (int i = 0 ; i < 10000; i++) {
        LOGI("Getting rotation: ");
        vector<double> rot = alc_getrotation(lls);
        LOGI("%lf %lf %lf\n", rot[0], rot[1], rot[2]);
    }
    LOGI("Disabling sensors: \n");
    alc_disablesensors(lls);

    alc_disconnect(lls);
    return 0;
}